# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################




from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import cache
from mio import common
from mio import eal
from mio import install
from mio import doctor
from tqdm import tqdm
from tqdm import trange
from threading import Thread
from multiprocessing.pool import ThreadPool
from threading import BoundedSemaphore
import threading
import time
import os
from datetime import datetime
import yaml
from yaml.loader import SafeLoader
import math
import re
import atexit
import tarfile




bwrap_ignore_list = [
    "xsim.dir", ".str", ".Xil", ".jou", ".log", ".wdb", ".vcd", ".log", ".sdb", ".rlx", ".pb", ".o", ".png", ".jpg",
    ".svg", ".vsdx", ".docx", ".xlsx", ".pptx", ".md", "sync", "workspace"
]

bwrap_ignore_dirs = [ ".git", ".svn" ]

regex_define_pattern  = "\+define\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
regex_plusarg_pattern = "\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
seconds_waited = 0
num_deps_to_compile = 0
sem = BoundedSemaphore(1)
pbar = None
est_time = 0
overtime = True


class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self,  *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()




class SimulationJob:
    """Simulation Job model"""
    
    def __init__(self, ip_str):
        self.vendor, self.ip = common.parse_dep(ip_str)
        self.simulator       = ""
        self.fsoc            = False
        self.compile         = False
        self.elaborate       = False
        self.simulate        = False
        self.test            = ""
        self.seed            = 0
        self.max_errors      = 0
        self.gui             = False
        self.verbosity       = ""
        self.waves           = False
        self.cov             = False
        self.dry_run         = False
        self.raw_args        = []
        self.cmp_args        = {}
        self.elab_args       = {}
        self.sim_args        = {}
        
        self.bwrap           = False
        self.bwrap_commands  = []
        self.bwrap_flists    = {}
        
        self.is_regression        = False
        self.regression_name      = ""
        self.regression_timestamp = ""
        
        self.timestamp_start    = ""
        self.timestamp_end      = ""
        self.filelist_path      = ""
        self.results_path       = ""
        self.results_dir_name   = ""
        self.cmp_log_file_path  = ""
        self.elab_log_file_path = ""
        self.sim_log_file_path  = ""
        




def main(sim_job):
    global est_time
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg(f"Starting simulation job: '{sim_job.vendor}/{sim_job.ip}'")
    if sim_job.vendor == "":
        ip = cache.get_anon_ip(sim_job.ip, True)
    else:
        ip = cache.get_ip(sim_job.vendor, sim_job.ip, True)
    if ip == None:
        common.fatal(f"Cannot find IP '{sim_job.vendor}/{sim_job.ip}'")
    ip_str = f"{ip.vendor}/{ip.name}"
    
    if sim_job.compile:
        if not ip.are_deps_installed():
            deps_to_install = ip.get_deps_to_install()
            if len(deps_to_install) > 0:
                install_deps_str = ""
                while (install_deps_str != "y") and (install_deps_str != "n"):
                    install_deps_str = common.prompt(f"{len(deps_to_install)} dependencies must first be installed.  Would you like to do so now? [y/n]").strip().lower()
                if install_deps_str == "y":
                    local_str = ""
                    while (local_str != "y") and (local_str != "n"):
                        local_str = common.prompt("Local install (vs. global)? [y/n]").strip().lower()
                    if local_str == "y":
                        global_install = False
                    else:
                        global_install = True
                    install.install_ip_and_deps(ip.vendor, ip.name, global_install)
                else:
                    common.fatal("Cannot continue without first installing '" + ip_str + "' IP dependencies.")
    
    if not doctor.check_simulator_executables(sim_job.simulator):
        common.fatal("Simulator '" + sim_job.simulator + "' not installed properly or environment variable missing")
    
    convert_cli_args_to_defines (sim_job)
    convert_cli_args_to_plusargs(sim_job)
    create_sim_directories      (sim_job)
    check_dependencies(ip)
    check_dut         (ip)
    
    eal.init_workspace(sim_job)
    
    fsoc_core_name = ""
    fsoc_core_flist_path = ""
    if ip.dut_ip_type == "fsoc":
        if sim_job.fsoc:
            common.info(f"Processing FuseSoC Core '{ip.dut_core.name}' ...")
            fsoc_core_flist_path = eal.invoke_fsoc(ip, ip.dut_core, sim_job)
            if not sim_job.compile and not sim_job.elaborate and not sim_job.simulate:
                common.info(f"Done.")
        fsoc_core_name = ip.dut_core
    elif ip.dut_ip_type != "":
        if ip.dut_ip_type.target_ip_model.sub_type == "vivado":
            dut_ip_str = f"{ip.dut.vendor}/{ip.dut.target_ip}"
            common.info(f"Compiling Vivado Project DUT IP '{dut_ip_str}' ...")
            est_time = get_ip_cmp_est(ip.dut_ip_type.target_ip_model, sim_job)
            if est_time > 0:
                pool = ThreadPool(processes=1)
                pool.apply_async(progress_bar)
                eal.cmp_vivado_project(ip.dut.target_ip_model, sim_job)
                fill_progress_bar()
                pool.terminate()
                pool.join()
            else:
                eal.cmp_vivado_project(ip.dut.target_ip_model, sim_job)
    
    if sim_job.compile and sim_job.elaborate:
        if sim_job.simulator == common.simulators_enum.VIVADO:
            common.info(f"Compiling {ip_str} ...")
            est_time = get_ip_cmp_est(ip, sim_job)
            if est_time > 0:
                pool = ThreadPool(processes=1)
                pool.apply_async(progress_bar)
                eal.cmp_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
                fill_progress_bar()
                pool.terminate()
                pool.join()
            else:
                eal.cmp_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
            common.info(f"Elaborating {ip_str} ...")
            est_time = get_ip_elab_est(ip, sim_job)
            if est_time > 0:
                pool = ThreadPool(processes=1)
                pool.apply_async(progress_bar)
                eal.elab_ip(ip, sim_job)
                fill_progress_bar()
                pool.terminate()
                pool.join()
            else:
                eal.elab_ip(ip, sim_job)
        else:
            common.info(f"Compiling+Elaborating {ip_str} ...")
            est_time = get_ip_gen_image_est(ip, sim_job)
            if est_time > 0:
                pool = ThreadPool(processes=1)
                pool.apply_async(progress_bar)
                eal.gen_image_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
                fill_progress_bar()
                pool.terminate()
                pool.join()
            else:
                eal.gen_image_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
    elif sim_job.compile:
        common.info(f"Compiling {ip_str} ...")
        est_time = get_ip_cmp_est(ip, sim_job)
        if est_time > 0:
            pool = ThreadPool(processes=1)
            pool.apply_async(progress_bar)
            eal.cmp_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
            fill_progress_bar()
            pool.terminate()
            pool.join()
        else:
            eal.cmp_ip(ip, sim_job, fsoc_core_name, fsoc_core_flist_path)
    elif sim_job.elaborate:
        common.info(f"Elaborating {ip_str} ...")
        est_time = get_ip_elab_est(ip, sim_job)
        if est_time > 0:
            pool = ThreadPool(processes=1)
            pool.apply_async(progress_bar)
            eal.elab_ip(ip, sim_job)
            fill_progress_bar()
            pool.terminate()
            pool.join()
        else:
            eal.elab_ip(ip, sim_job)
    
    if not sim_job.is_regression:
        if sim_job.simulate:
            common.banner(f"Simulating {ip_str} ...")
            if sim_job.gui and sim_job.simulator == common.simulators_enum.METRICS:
                common.warning("The Metrics Cloud Simulator does not support GUI mode")
            eal.simulate(ip, sim_job)
    
    if not sim_job.is_regression:
        if sim_job.compile and sim_job.elaborate:
            if sim_job.simulator == common.simulators_enum.VIVADO:
                print_end_of_compilation_message(ip, sim_job)
                print_end_of_elaboration_message(ip, sim_job)
            else:
                print_end_of_gen_image_message(ip, sim_job)
        elif sim_job.compile:
            print_end_of_compilation_message(ip, sim_job)
        elif sim_job.elaborate:
            print_end_of_elaboration_message(ip, sim_job)
        if sim_job.simulate:
            print_end_of_simulation_message(ip, sim_job)
    
    if sim_job.bwrap:
        bubble_wrap(sim_job)



def convert_cli_args_to_defines(sim_job):
    defines = {}
    args = sim_job.raw_args
    if args != None:
        if type(args) is list:
            for item in args:
                regex = "\+define\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                defines[arg_name] = arg_value
                common.dbg("Added define '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if result:
                    define_name = result.group(1)
                    if len(result.groups()) > 1:
                        define_value = result.group(2)
                        if define_value == None:
                            arg_value = ""
                    else:
                        define_value = ""
                    defines[define_name] = define_value
                    common.dbg("Added define '" + define_name + "' with value '" + define_value + "' to list")
    sim_job.cmp_args = defines


def convert_cli_args_to_plusargs(sim_job):
    define_regex = "\+define\+(\w+)(?:\=(\w+))?"
    plus_args = {}
    args = sim_job.raw_args
    if args != None:
        if type(args) is list:
            for item in args:
                result = re.match(define_regex, item)
                if result:
                    continue
                regex = "\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                plus_args[arg_name] = arg_value
                common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if not result:
                    result = re.match(regex_plusarg_pattern, arg)
                    if result:
                        arg_name = result.group(1)
                        if len(result.groups()) > 1:
                            arg_value = result.group(2)
                            if arg_value == None:
                                arg_value = ""
                        else:
                            arg_value = ""
                        plus_args[arg_name] = arg_value
                        common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
    sim_job.sim_args = plus_args


def create_sim_directories(sim_job):
    common.dbg("Creating sim directories")
    common.create_dir(cfg.sim_dir)
    common.create_dir(cfg.sim_output_dir)
    common.create_dir(cfg.sim_output_dir + "/viv"                     )
    common.create_dir(cfg.sim_output_dir + "/viv/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/viv/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/viv/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/viv/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/viv/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/vcs"                     )
    common.create_dir(cfg.sim_output_dir + "/vcs/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/vcs/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/vcs/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/vcs/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/vcs/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/xcl"                     )
    common.create_dir(cfg.sim_output_dir + "/xcl/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/xcl/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/xcl/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/xcl/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/xcl/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/qst"                     )
    common.create_dir(cfg.sim_output_dir + "/qst/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/qst/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/qst/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/qst/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/qst/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/riv"                     )
    common.create_dir(cfg.sim_output_dir + "/riv/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/riv/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/riv/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/riv/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/riv/so_libs"             )
    common.create_dir(cfg.sim_dir + "/cmp")
    common.create_dir(cfg.regr_results_dir)
    common.create_dir(cfg.sim_results_dir)


def check_dependencies(ip):
    for dep in ip.dependencies:
        #common.dbg(f"Checking dependency '{dep.vendor}/{dep.target_ip}'")
        found_ip = cache.get_ip(dep.vendor, dep.target_ip)
        if found_ip == None:
            common.fatal(f"Could not find IP dependency '{dep.vendor}/{dep.target_ip}'")
        else:
            check_dependencies(found_ip)


def check_dut(ip):
    if ip.dut_ip_type == "fsoc":
        pass
    else:
        if ip.dut != None:
            if ip.dut.target_ip_model == None:
                common.fatal(f"Did not resolve DUT dependency ('{ip.dut.vendor}/{ip.dut.target_ip}')!")




def bubble_wrap(sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    run_script_path = f"{cfg.sim_dir}/run.sh"
    readme_path = f"{cfg.sim_dir}/README.txt"
    
    if sim_job.simulator == common.simulators_enum.VIVADO:
        bin_text = "MIO_VIVADO_HOME"
        bin_string = "/tools/vivado/bin"
    if sim_job.simulator == common.simulators_enum.METRICS:
        bin_text = "MIO_METRICS_HOME"
        bin_string = "/usr/local/bin"
    if sim_job.simulator == common.simulators_enum.VCS:
        bin_text = "MIO_VCS_HOME"
        bin_string = "/tools/vcs/bin"
    if sim_job.simulator == common.simulators_enum.QUESTA:
        bin_text = "MIO_QUESTA_HOME"
        bin_string = "/tools/questa/bin"
    if sim_job.simulator == common.simulators_enum.XCELIUM:
        bin_text = "MIO_XCELIUM_HOME"
        bin_string = "/tools/xcelium/bin"
    if sim_job.simulator == common.simulators_enum.RIVIERA:
        bin_text = "MIO_RIVIERA_HOME"
        bin_string = "/tools/riviera/bin"
    
    try:
        with open(run_script_path, 'w') as run_script_file:
            run_script_file.write(f"export PROJECT_ROOT_DIR=$(pwd)/..\n\n")
            for flist in sim_job.bwrap_flists:
                run_script_file.write(f"export {flist}={sim_job.bwrap_flists[flist]}\n")
            run_script_file.write("\n\n\n\n")
            for cmd in sim_job.bwrap_commands:
                cmd = cmd.replace(cfg.project_dir , "${PROJECT_ROOT_DIR}")
                cmd = cmd.replace("-f .mio/"      , "-f ${PROJECT_ROOT_DIR}/.mio/")
                cmd = cmd.replace(cfg.vivado_home , "${MIO_VIVADO_HOME}" )
                cmd = cmd.replace(cfg.metrics_home, "${MIO_METRICS_HOME}")
                cmd = cmd.replace(cfg.vcs_home    , "${MIO_VCS_HOME}"    )
                cmd = cmd.replace(cfg.xcelium_home, "${MIO_XCELIUM_HOME}")
                cmd = cmd.replace(cfg.questa_home , "${MIO_QUESTA_HOME}" )
                cmd = cmd.replace(cfg.riviera_home, "${MIO_RIVIERA_HOME}")
                run_script_file.write(f"{cmd}\n\n")
        common.info(f"Wrote {run_script_path}")
        with open(readme_path, 'w') as readme_file:
            readme_file.write(f"1. Set ${bin_text}.  Ex: export {bin_text}={bin_string}\n")
            readme_file.write(f"2. Run: bash ./run.sh\n")
        common.info(f"Wrote {readme_path}")
        
        tarball_filename = f"{sim_job.ip}.{sim_job.test}.{sim_job.seed}.{sim_str}.tgz"
        tarball_path = f"{cfg.project_dir}/../{tarball_filename}"
        common.info(f"Writing {tarball_path} ...")
        with tarfile.open(tarball_path, "w:gz") as tar:
            files = os.listdir(cfg.project_dir)
            for file in files:
                file_path = f"{cfg.project_dir}/{file}"
                if os.path.isdir(file_path):
                    if file in bwrap_ignore_dirs:
                        continue
                common.dbg(f"Compressing {file}")
                tar.add(file_path, filter=bwrap_exclude, arcname=file)
        common.info("Done")
        
    except Exception as e:
        common.fatal(f"Failed to create bubble-wrap tarball: {e}")



def bwrap_exclude(tar_info):
    for regex in bwrap_ignore_list:
        if tar_info.name.endswith(regex):
            return None
    return tar_info




def print_end_of_gen_image_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.simulate:
        common.info("************************************************************************************************************************")
        common.info("* Compilation+Elaboration results: " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".log")
    else:
        common.info("************************************************************************************************************************")
        common.info("* Compilation+Elaboration results:")
        common.info("************************************************************************************************************************")
        common.info("  emacs " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".log &")
        common.info("  gvim  " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".log &")
        common.info("  vim   " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".log")
        common.info("")


def print_end_of_compilation_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.simulate or sim_job.elaborate:
        common.info("************************************************************************************************************************")
        common.info("* Compilation results: " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log")
    else:
        common.info("************************************************************************************************************************")
        common.info("* Compilation results:")
        common.info("************************************************************************************************************************")
        common.info("  emacs " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log &")
        common.info("  gvim  " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log &")
        common.info("  vim   " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log")
        common.info("")


def print_end_of_elaboration_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.simulate:
        common.info("************************************************************************************************************************")
        common.info("* Elaboration results: " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".elab.log")
    else:
        common.info("************************************************************************************************************************")
        common.info("* Elaboration results")
        common.info("************************************************************************************************************************")
        common.info("  emacs " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".elab.log &")
        common.info("  gvim  " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".elab.log &")
        common.info("  vim   " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".elab.log")
        common.info("")


def print_end_of_simulation_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    results_path = sim_job.results_path
    common.info("************************************************************************************************************************")
    common.info("* Simulation results")
    common.info("************************************************************************************************************************")
    if (sim_job.waves):
        if sim_job.simulator == common.simulators_enum.VIVADO:
            common.info("* Waveforms: $MIO_VIVADO_HOME/xsim -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.VCS:
            common.info("* Waveforms: $MIO_VCS_HOME/dve -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.METRICS:
            common.info(f"* Waveforms: gtkwave {results_path}/waves.vcd &")
        elif sim_job.simulator == common.simulators_enum.XCELIUM:
            common.info("* Waveforms: $MIO_XCELIUM_HOME/simvision -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.QUESTA:
            common.info("* Waveforms: $MIO_QUESTA_HOME/visualizer -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.RIVIERA:
            common.info("* Waveforms: $MIO_RIVIERA_HOME/??? -gui " + results_path + "/waves.wdb &")
        common.info("")
    common.info("* Main log: emacs " + results_path + "/sim.log &")
    common.info("            gvim  " + results_path + "/sim.log &")
    common.info("            vim   " + results_path + "/sim.log")
    common.info("")
    common.info("* Test result dir : pushd " + results_path)
    common.info("* UVMx logs       : pushd " + results_path + "/uvmx_logs")
    common.info("")



def get_ip_cmp_est(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    est_time = 0
    if ip_str in cfg.job_history:
        if 'compilation' in cfg.job_history[ip_str]:
            for job in cfg.job_history[ip_str]['compilation']:
                duration = job['duration']
                est_time += int(duration)
            est_time = math.ceil(est_time / len(cfg.job_history[ip_str]['compilation']))
    return est_time



def get_ip_elab_est(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    est_time = 0
    if ip_str in cfg.job_history:
        if 'elaboration' in cfg.job_history[ip_str]:
            for job in cfg.job_history[ip_str]['elaboration']:
                duration = job['duration']
                est_time += int(duration)
            est_time = math.ceil(est_time / len(cfg.job_history[ip_str]['elaboration']))
    return est_time



def get_ip_gen_image_est(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    est_time = 0
    if ip_str in cfg.job_history:
        if 'gen-image' in cfg.job_history[ip_str]:
            for job in cfg.job_history[ip_str]['gen-image']:
                duration = job['duration']
                est_time += int(duration)
            est_time = math.ceil(est_time / len(cfg.job_history[ip_str]['gen-image']))
    return est_time



def kill_progress_bar():
    global pbar
    if pbar:
        pbar.close()
atexit.register(kill_progress_bar)



def fill_progress_bar():
    global pbar
    global seconds_waited
    global est_time
    global overtime
    if pbar:
        pbar.update(est_time-seconds_waited)
        pbar.close()
        overtime = False



def progress_bar():
    global pbar
    global seconds_waited
    global est_time
    global overtime
    with tqdm(total=est_time, bar_format="{l_bar} {bar} {elapsed}<{remaining}") as pbar:
        for seconds_waited in range(est_time):
            time.sleep(1)
            pbar.update(1)
    if overtime:
        common.info(f"Job taking longer than average ({est_time} seconds)")






# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cfg
from mio import cache
from mio import user

import yaml
from yaml import SafeLoader

import requests
import getpass
import tarfile
import json
from pathlib import Path
from base64 import b64decode
import os
import shutil
from tqdm import tqdm


base_url          = "https://mooreio.com"#"http://localhost:8080"
jwt_endpoint      = base_url + "/api/authenticate"
ips_endpoint      = base_url + "/api/ips"
versions_endpoint = base_url + "/api/versions"
version_endpoint  = base_url + "/api/version/"
licenses_endpoint = base_url + "/api/licenses"
headers = {}


def install_ip(vendor, name, global_install, username="", password=""):
    if global_install:
        location = cfg.user_global_ips_path
    else:
        location = cfg.dependencies_path
    common.dbg(f"Installing IP '{vendor}/{name}' under '{location}'")
    query = {'name': name}
    
    token = user.get_token()
    if token == None:
        token = user.login(username, password)
    
    headers = {'Authorization':'Bearer ' + token}
    response = requests.get(ips_endpoint + "?page=0&size=1000000", params=query, headers=headers)
    json_data = response.json()
    common.dbg(f"Response from Moore.io IP Marketplace: '{response}'")
    payload = None
    found_payload = False
    version_str = ""
    for ip in json_data:
        if ip['name'] == name:
            ip_id = ip['id']
            common.dbg("Found IP! name=" + name + " id=" + str(ip_id))
            license_type = ip['licenseType']
            if license_type == "FREE_OPEN_SOURCE":
                versions = requests.get(versions_endpoint + "?page=0&size=1000000", headers=headers).json()
                for version in versions:
                    if version['ip']['id'] == ip_id:
                        version_str = version['semver']
                        common.dbg("Found IP version on server: " + name + " v" + version_str)
                        payload = version['publicPayload']
                        found_payload = True
                        filename = Path(location + "/" + vendor + "__" + name + '.tgz')
                        filename.write_bytes(b64decode(payload))
                        tar = tarfile.open(filename, "r:gz")
                        ip_destination_path = f"{location}/{vendor}__{name}"
                        common.remove_dir(ip_destination_path)
                        common.create_dir(ip_destination_path)
                        tar.extractall(ip_destination_path)
                        tar.close()
                        common.remove_file(filename)
                        break
            if license_type == "COMMERCIAL":
                licenses = requests.get(licenses_endpoint + "?page=0&size=1000000", headers=headers).json()
                for license in licenses:
                    if license['targetIp']['id'] == ip_id:
                        version_str = license['version']['semver']
                        common.dbg("Found IP license on server: " + name)
                        payload = license['payload']
                        found_payload = True
                        filename = Path(location + "/" + vendor + "__" + name + '.tgz')
                        filename.write_bytes(b64decode(payload))
                        tar = tarfile.open(filename, "r:gz")
                        ip_destination_path = f"{location}/{vendor}__{name}"
                        common.remove_dir(ip_destination_path)
                        common.create_dir(ip_destination_path)
                        tar.extractall(ip_destination_path)
                        tar.close()
                        common.remove_file(filename)
    if not found_payload:
        common.fatal(f"Could not find IP '{vendor}/{name}' on Moore.io IP Marketplace")
    
    return version_str


def install_ip_dep_list(ip, ip_list, global_install, username="", password=""):
    versions = {}
    if len(ip_list) > 0:
        #with alive_bar(len(ip_list), bar = 'smooth') as bar:
        with tqdm(ip_list) as pbar:
            for dep_ip in ip_list:
                vendor, name = common.parse_dep(dep_ip)
                ip_str = f"{vendor}/{name}"
                pbar.set_description(ip_str)
                versions[ip_str] = install_ip(vendor, name, global_install, username="", password="")
                pbar.update(1)
        cache.scan_and_load_ip_metadata()
        update_lock_file(ip, versions)


def install_ip_list(ip, ip_list, global_install, username="", password=""):
    versions = {}
    if len(ip_list) > 0:
        #with alive_bar(len(ip_list), bar = 'smooth') as bar:
        with tqdm(ip_list) as pbar:
            for dep_ip in ip_list:
                vendor = ip.vendor
                name = ip.name
                ip_str = f"{vendor}/{name}"
                pbar.set_description(ip_str)
                versions[ip_str] = install_ip(vendor, name, global_install, username="", password="")
                pbar.update(1)
        cache.scan_and_load_ip_metadata()
        update_lock_file(ip, versions)


def install_ip_and_deps(vendor, name, global_install, username="", password=""):
    ip = {}
    if vendor == "":
        ip = cache.get_anon_ip(name)
    else:
        ip = cache.get_ip(vendor, name)
    if ip == None:
        common.banner(f"Installing IP '{vendor}/{name}'")
        install_ip(vendor, name, global_install, username, password)
        cache.scan_and_load_ip_metadata()
        ip = cache.get_ip(vendor, name)
    
    ip.delete_dependencies()
    cache.scan_and_load_ip_metadata()
    if vendor == "":
        ip = cache.get_anon_ip(name)
    else:
        ip = cache.get_ip(vendor, name)
    
    #deps_to_install = ip.get_ordered_install_deps()
    #common.info(f"Installing {len(deps_to_install)} dependencies")
    #install_ip_list(ip, deps_to_install, global_install, username, password)
    #cache.scan_and_load_ip_metadata()
    
    deps_to_install = ip.get_deps_to_install()
    while len(deps_to_install) > 0:
        common.info(f"Installing {len(deps_to_install)} dependencies")
        install_ip_dep_list(ip, deps_to_install, global_install, username, password)
        cache.scan_and_load_ip_metadata()
        deps_to_install = ip.get_deps_to_install()


def update_lock_file(ip, versions={}):
    ip_file_path   = ip.path + "/ip.yml"
    lock_file_path = ip.path + "/ip.lock.yml"
    try:
        common.copy_file(ip_file_path, lock_file_path)
        with open(lock_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if 'dependencies' in ymlr:
                if type(ymlr['dependencies']) is dict:
                    for dep in ymlr['dependencies']:
                        if dep in versions:
                            ymlr['dependencies'][dep] = versions[dep]
                        else:
                            vendor, name = common.parse_dep(dep.strip().lower().replace("@", ""))
                            if vendor == "":
                                dep_ip = cache.get_anon_ip(name, True)
                            else:
                                dep_ip = cache.get_ip(vendor, name, True)
                            if dep_ip.is_local:
                                ymlr['dependencies'][dep] = '@local'
                            else:
                                ymlr['dependencies'][dep] = dep_ip.version
        with open(lock_file_path, 'w') as yaml_file_write:
            yaml.dump(ymlr, yaml_file_write)
    except Exception as e:
        common.fatal(f"Failed to update '{lock_file_path}': {e}")




# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cfg
from mio import cache
from mio import install
from mio import eal
from mio import user

import yaml
from yaml import SafeLoader

import glob
import requests
import getpass
import tarfile
import json
from pathlib import Path
from base64 import b64encode
import os
import shutil
from tqdm import tqdm
import fileinput


base_url          = "https://mooreio.com"#"http://localhost:8080"
jwt_endpoint      = base_url + "/api/authenticate"
ips_endpoint      = base_url + "/api/ips"
versions_endpoint = base_url + "/api/versions"
licenses_endpoint = base_url + "/api/licenses"
headers = {}

ignore_files = [".gitignore", "ip.lock.yml"]


def do_publish_ip(ip_model, token, org):
    name = ip_model.name
    org  = ip_model.vendor
    location = ip_model.path
    common.dbg(f"Publishing IP '{name}' under '{location}'")
    query = {'name': name}
    auth_headers = {
        'Authorization'    : f"Bearer {token}",
        'Accept-Encoding'  : 'gzip, deflate, br',
        'Content-Type'     : 'application/json; charset="UTF-8"'
    }
    
    response = requests.get(ips_endpoint + "?page=0&size=1000000", params=query, headers=auth_headers)
    if response.status_code != 200:
        common.fatal(f"Failed to reach Moore.io IP Marketplace: '{response.status_code}' - {response.reason}")
    json_data = response.json()
    payload = None
    found_ip = False
    
    for ip in json_data:
        if ip['name'] == name:
            ip_id = ip['id']
            common.dbg("Found IP! name=" + name + " id=" + str(ip_id))
            license_type = ip['licenseType']
            if license_type == "FREE_OPEN_SOURCE":
                versions = requests.get(versions_endpoint + "?page=0&size=1000000", headers=auth_headers).json()
                for version in versions:
                    if version['ip']['id'] == ip_id:
                        common.dbg("Found IP version on server: " + name + " v" + version['semver'])
                        found_ip = True
                        update_shrinkwrap_file(ip_model, False)
                        tarball_path = package_ip(ip_model, False)
                        try:
                            with open(tarball_path,'rb') as f:
                                payload = b64encode(f.read())
                        except Exception as e:
                            common.fatal(f"Could not open '{tarball_path}': {e}")
                        url = versions_endpoint + f"/{version['id']}"
                        data = {
                            "id"                       : version['id']     ,
                            "publicPayload"            : str(payload)[2:-1],
                            "publicPayloadContentType" : "application/x-compressed"
                        }
                        common.dbg(f"Creating PATCH request to '{url}'")
                        try:
                            response = requests.patch(url=url, headers=auth_headers, json=data, verify=False)
                        except Exception as e:
                            common.fatal(f"Failed to publish IP '{name}': {e}")
                        if response.status_code != 200:
                            common.fatal(f"Failed to publish IP '{name}': '{response.status_code}' - {response.reason}")
                        break
            if license_type == "COMMERCIAL":
                if org == "" or org == None:
                    common.fatal(f"Organization name must be specified for commercial IPs")
                licenses = requests.get(licenses_endpoint + "?page=0&size=1000000", headers=auth_headers).json()
                for license in licenses:
                    if license['targetIp']['id'] == ip_id:
                        common.dbg("Found IP license on server: " + name)
                        if license['owner']['name'].strip().lower() == org.strip().lower():
                            common.dbg(f"Found IP '{name}' license on server for '{org}'")
                            found_ip = True
                            ip_key = license['key']
                            org_id = license['owner']['id']
                            update_shrinkwrap_file(ip_model, True)
                            tarball_path = package_ip(ip_model, True, "", True, org_id, ip_id, ip_key)
                            try:
                                with open(tarball_path,'rb') as f:
                                    payload = b64encode(f.read())
                            except Exception as e:
                                common.fatal(f"Could not open '{tarball_path}': {e}")
                            url = licenses_endpoint + f"/{license['id']}"
                            data = {
                                "id"                 : license['id']     ,
                                "payload"            : str(payload)[2:-1],
                                "payloadContentType" : "application/x-compressed"
                            }
                            common.dbg(f"Creating PATCH request to '{url}'")
                            try:
                                response = requests.patch(url=url, headers=auth_headers, json=data, verify=False)
                            except Exception as e:
                                common.fatal(f"Failed to publish IP '{name}': {e}")
                            if response.status_code != 200:
                                common.fatal(f"Failed to publish IP '{name}': '{response.status_code}' - {response.reason}")
                            break
    if found_ip:
        common.remove_file(tarball_path)
        common.info(f"Published IP '{name}' successfully")
    else:
        common.fatal(f"Could not find IP '{name}' on Moore.io IP Marketplace")


def cli_package_ip(ip_str, destination, no_tarball):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name)
    else:
        ip = cache.get_ip(vendor, name)
    create_tarball = not no_tarball
    common.banner(f"Packaging IP '{ip.vendor}/{ip.name}' ...")
    update_shrinkwrap_file(ip, ip.is_licensed)
    path = package_ip(ip, ip.is_licensed, str(destination.resolve()), create_tarball)
    if create_tarball:
        common.info(f"Compressed archive created at {path}")
    else:
        if ip.is_licensed:
            common.info(f"Encrypted IP code output to {path}")
        else:
            common.info(f"IP code output to {path}")


def package_ip(ip, encrypt=False, destination="", create_tarball=True, org_id="", ip_id="", ip_key=""):
    ip_name = ip.name
    ip_str = f"{ip.vendor}/{ip.name}"
    tarball_path = cfg.temp_path + "/" + ip_name + ".tgz"
    location = ip.path
    scripts_src_dir      = ip.scripts_path
    docs_src_dir         = ip.docs_path
    examples_src_dir     = ip.examples_path
    src_dir              = ip.src_path
    ip_src_directories   = ip.hdl_src_directories
    common.dbg(f"Creating tarball '{location}' -> '{tarball_path}'")
    temp_location        = cfg.temp_path + "/" + ip_name
    ip_yml_file_path     = temp_location + "/ip.yml"
    shrinkwrap_file_path = temp_location + "/ip.shrinkwrap.yml"
    lock_file_path       = temp_location + "/ip.lock.yml"
    
    if destination != "":
        tarball_path = destination + "/" + ip_name + ".tgz"
    
    if encrypt:
        sims_supported = 0
        if ip.simulators_supported[common.simulators_enum.VIVADO] != "":
            sims_supported = sims_supported + 1
        if ip.simulators_supported[common.simulators_enum.METRICS] != "":
            sims_supported = sims_supported + 1
        if ip.simulators_supported[common.simulators_enum.VCS] != "":
            sims_supported = sims_supported + 1
        if ip.simulators_supported[common.simulators_enum.XCELIUM] != "":
            sims_supported = sims_supported + 1
        if ip.simulators_supported[common.simulators_enum.QUESTA] != "":
            sims_supported = sims_supported + 1
        if ip.simulators_supported[common.simulators_enum.RIVIERA] != "":
            sims_supported = sims_supported + 1
        if sims_supported == 0:
            common.fatal(f"IP '{ip_str}' does not have any simulators supported!")
        common.dbg(f"Encrypting IP '{ip_str}'")
        vendor = ip.vendor
        common.copy_directory(location, temp_location)
        common.copy_file(shrinkwrap_file_path, ip_yml_file_path)
        common.remove_file(shrinkwrap_file_path)
        common.remove_file(lock_file_path)
        src_dir = temp_location + "/" + ip.src_path
        insert_key_checks(src_dir, ip, org_id, ip_id, ip_key)
        
        if ip.simulators_supported[common.simulators_enum.VIVADO] != "":
            viv_src_dir = src_dir + ".viv"
            common.copy_directory(src_dir, viv_src_dir)
            eal.encrypt_tree(ip_name, viv_src_dir, "viv")
        
        if ip.simulators_supported[common.simulators_enum.VCS] != "":
            vcs_src_dir = src_dir + ".vcs"
            common.copy_directory(src_dir, vcs_src_dir)
            eal.encrypt_tree(ip_name, vcs_src_dir, "vcs")
        
        if ip.simulators_supported[common.simulators_enum.METRICS] != "":
            mtr_src_dir = src_dir + ".mdc"
            common.copy_directory(src_dir, mtr_src_dir)
            eal.encrypt_tree(ip_name, mtr_src_dir, "mdc")
        
        if ip.simulators_supported[common.simulators_enum.QUESTA] != "":
            qst_src_dir = src_dir + ".qst"
            common.copy_directory(src_dir, qst_src_dir)
            eal.encrypt_tree(ip_name, qst_src_dir, "qst")
        
        if ip.simulators_supported[common.simulators_enum.XCELIUM] != "":
            xcl_src_dir = src_dir + ".xcl"
            common.copy_directory(src_dir, xcl_src_dir)
            eal.encrypt_tree(ip_name, xcl_src_dir, "xcl")
        
        if ip.simulators_supported[common.simulators_enum.RIVIERA] != "":
            riv_src_dir = src_dir + ".riv"
            common.copy_directory(src_dir, riv_src_dir)
            eal.encrypt_tree(ip_name, riv_src_dir, "riv")
        
        if create_tarball:
            try:
                with tarfile.open(tarball_path, "w:gz") as tar:
                    if os.path.exists(temp_location + "/" + ip.scripts_path):
                        tar.add(temp_location + "/" + scripts_src_dir , arcname=scripts_src_dir)
                    if os.path.exists(temp_location + "/" + docs_src_dir):
                        tar.add(temp_location + "/" + docs_src_dir, arcname=docs_src_dir)
                    if os.path.exists(temp_location + "/" + examples_src_dir):
                        tar.add(temp_location + "/" + examples_src_dir, arcname=examples_src_dir)
                    
                    if ip.simulators_supported[common.simulators_enum.VIVADO] != "":
                        tar.add(viv_src_dir, arcname="src.viv")
                    
                    if ip.simulators_supported[common.simulators_enum.VCS] != "":
                        tar.add(vcs_src_dir, arcname="src.vcs")
                    
                    if ip.simulators_supported[common.simulators_enum.METRICS] != "":
                        tar.add(mtr_src_dir, arcname="src.mdc")
                    
                    if ip.simulators_supported[common.simulators_enum.QUESTA] != "":
                        tar.add(qst_src_dir, arcname="src.qst")
                    
                    if ip.simulators_supported[common.simulators_enum.XCELIUM] != "":
                        tar.add(xcl_src_dir, arcname="src.xcl")
                    
                    if ip.simulators_supported[common.simulators_enum.RIVIERA] != "":
                        tar.add(riv_src_dir, arcname="src.riv")
                    
                    for path in os.listdir(temp_location):
                        file_path = os.path.join(temp_location, path)
                        if os.path.isfile(file_path):
                            if path not in ignore_files:
                                tar.add(file_path, arcname=path)
                common.remove_dir(temp_location)
            except Exception as e:
                common.fatal(f"Failed to create tarball for IP '{ip_str}': {e}")
    
    else:
        if create_tarball:
            try:
                common.copy_directory(location, temp_location)
                common.copy_file(shrinkwrap_file_path, ip_yml_file_path)
                common.remove_file(shrinkwrap_file_path)
                common.remove_file(lock_file_path)
                with tarfile.open(tarball_path, "w:gz") as tar:
                    if os.path.exists(temp_location + "/" + scripts_src_dir):
                        tar.add(temp_location + "/" + scripts_src_dir , arcname=scripts_src_dir)
                    if os.path.exists(temp_location + "/" + docs_src_dir):
                        tar.add(temp_location + "/" + docs_src_dir, arcname=docs_src_dir)
                    if os.path.exists(temp_location + "/" + examples_src_dir):
                        tar.add(temp_location + "/" + examples_src_dir, arcname=examples_src_dir)
                    tar.add(temp_location + "/" + src_dir, arcname=src_dir)
                    for path in os.listdir(temp_location):
                        file_path = os.path.join(temp_location, path)
                        if os.path.isfile(file_path):
                            if path not in ignore_files:
                                tar.add(file_path, arcname=path)
                    common.remove_dir(temp_location)
            except Exception as e:
                common.fatal(f"Failed to create tarball for IP '{ip_str}': {e}")
        else:
            common.copy_directory(location, destination + "/" + ip.name)
    
    if create_tarball:
        return tarball_path
    else:
        return destination + "/" + ip.name


def publish_ip(ip_str, username="", password="", org=""):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name, True)
    else:
        ip = cache.get_ip(vendor, name, True)
    target_name = f"{ip.vendor}/{ip.name}"
    if not ip.is_local:
        common.fatal(f"IP '{target_name}' is not local.")
    if (username == "") or (username == None):
        token = user.login()
    else:
        token = user.login(username, password, True)
    do_publish_ip(ip, token, org)


def insert_key_checks(location, ip, org_id, ip_id, ip_key):
    ip_name = ip.name
    ip_vendor = ip.vendor
    text_to_search = "`MIO_LICENSE_CHECK"
    for file in glob.iglob(location + '**/**', recursive=True):
        file_path = os.path.join(location, file)
        if (file_path[-2:] == ".v") or (file_path[-3:] == ".vh") or (file_path[-3:] == ".sv") or (file_path[-4:] == ".svh"):  # TODO Add support for VHDL files
            try:
                with open(file_path, "r") as file_o:
                    replacement_text = f"uvml_mio_lic_pkg::uvml_mio_lic_server_c::check_key(\"{ip_vendor}/{ip_name}/{file}\", \"{ip_id}\", \"{org_id}\", \"{ip_key}\");"
                    content = file_o.read().replace(text_to_search, replacement_text)
                file_o = open(file_path, "w")
                file_o.write(content)
                file_o.close()
            except Exception as e:
                common.fatal(f"Error while processing file '{file_path}' for key check insertion: {e}")


def update_shrinkwrap_file(ip, encrypted=False):
    install.update_lock_file(ip)
    lock_file_path       = ip.path + "/ip.lock.yml"
    shrinkwrap_file_path = ip.path + "/ip.shrinkwrap.yml"
    try:
        common.copy_file(lock_file_path, shrinkwrap_file_path)
        with open(shrinkwrap_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            ymlr['hdl-src']['encrypted'] = encrypted
        with open(shrinkwrap_file_path, 'w') as yaml_file_write:
            yaml.dump(ymlr, yaml_file_write)
    except Exception as e:
        common.fatal(f"Failed to update '{shrinkwrap_file_path}': {e}")



# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import sim
from mio import regr
from mio import install
from mio import publish
from mio import cache
from mio import new
from mio import common
from mio import init
from mio import main
from mio import help_text
from mio import user
from mio import doctor

import sys
import argparse
import os
import subprocess
import shutil
import yaml
from datetime import datetime
from yaml.loader import SafeLoader
import re
import random
import pathlib


uvm_levels      = ["none","low","medium","high","debug"]
simulators      = ["viv","mdc","vcs","xcl","qst","riv"]
commands        = ["clean", "cov", "doctor", "dox", "init" ,"install", "login", "new", "package", "publish", "regr", "results", "sim", "!"]
repeat_commands = ["sim"]


def main():
    parser       = build_parser()
    cli_args     = parser.parse_args()
    cfg.dbg      = cli_args.dbg
    common.dbg("CLI arguments: " + str(cli_args))
    cfg.cli_args = cli_args
    
    if cli_args.help:
        print_help_text()
    if cli_args.version:
        print_version()
    
    if cli_args.command == "help":
        print_cmd_help_text(cli_args)
        common.exit(False)
    
    if cli_args.wd != None:
        cfg.set_pwd(cli_args.wd)
    else:
        cfg.pwd = os.getcwd()
    
    user.load_user_data()
    in_project = cfg.find_project_descriptor()
    
    if cli_args.command == "doctor":
        doctor.main()
        common.exit()
    
    if not in_project:
        if cli_args.command != "init":
            common.fatal("Could not find 'mio.toml' project file")
        elif cli_args.command == "init":
            init.new_project(cfg.pwd)
            common.exit(False)
    
    cfg.load_tree()
    cfg.load_configuration()
    common.create_common_files()
    log_cli_args_to_disk()
    cache.scan_and_load_ip_metadata()
    
    if cli_args.command == 'init':
        init.new_ip(cfg.pwd)
        common.exit()
    if cli_args.command == 'new':
        new.menu(cfg.cli_args.template)
        common.exit()
    if cli_args.command == 'install':
        if (cli_args.username != None) and (cli_args.username != ""):
            if (cli_args.password == None) or (cli_args.password == ""):
                common.fatal("Must specify both username AND password")
            else:
                user.login(cli_args.username, cli_args.password, True)
        vendor, name = common.parse_dep(cli_args.ip.lower())
        install.install_ip_and_deps(vendor, name, cli_args.is_global, cli_args.username, cli_args.password)
        common.exit()
    if cli_args.command == 'login':
        if (cli_args.username != None) and (cli_args.username != ""):
            if (cli_args.password == None) or (cli_args.password == ""):
                common.fatal("Must specify both username AND password")
        user.login(cli_args.username, cli_args.password, True)
        common.exit()
    if cli_args.command == 'package':
        cache.check_ip_str(cli_args.ip.lower())
        publish.cli_package_ip(cli_args.ip.lower(), cli_args.dest, cli_args.no_tarball)
        common.exit()
    if cli_args.command == 'publish':
        if (cli_args.username != None) and (cli_args.username != ""):
            if (cli_args.password == None) or (cli_args.password == ""):
                common.fatal("Must specify both username AND password")
        cache.check_ip_str(cli_args.ip.lower())
        publish.publish_ip(cli_args.ip.lower(), cli_args.username, cli_args.password, cli_args.org)
        common.exit()
    if cli_args.command == 'dox':
        cache.check_ip_str(cli_args.ip.lower())
        dox.gen_doxygen(cli_args.ip.lower())
        common.exit()
    if cli_args.command == 'cov':
        cache.check_ip_str(cli_args.ip.lower())
        common.banner(f"Generating coverage report for '{cli_args.ip.lower()}'")
        report_path = cov.gen_cov_report(cli_args.ip.lower())
        common.info(f"Coverage report: `pushd {report_path}`")
        common.exit()
    if cli_args.command == 'results':
        cache.check_ip_str(cli_args.ip.lower())
        common.banner(f"Parsing simulation results for '{cli_args.ip.lower()}'")
        regr_results = results.main(cli_args.ip.lower(), cli_args.filename)
        common.info(f"HTML Report: '{regr_results.html_report_path}'")
        common.info(f"Jenkins XML: '{regr_results.xml_report_path}'")
        common.exit()
    if cli_args.command == 'clean':
        cache.check_ip_str(cli_args.ip.lower())
        #clean.main(cli_args.ip.lower(), cli_args.deep)
        clean.main(cli_args.ip.lower())
        common.exit()
    if cli_args.command == 'sim':
        sim_job = create_sim_job(cli_args)
        cache.check_ip(sim_job.vendor, sim_job.ip)
        user.login()
        sim.main(sim_job)
        common.exit()
    if cli_args.command == '!':
        sim_job = create_repeat_sim_job(cli_args)
        cache.check_ip(sim_job.vendor, sim_job.ip)
        user.login()
        sim.main(sim_job)
        common.exit()
    if cli_args.command == 'regr':
        cache.check_ip_str(cli_args.ip.lower())
        regr.main(cli_args.ip.lower(), cli_args.regr.lower(), cli_args.app, cli_args.dry)
        common.exit()
    
    common.fatal("No command specified")


def build_parser():
    parser = argparse.ArgumentParser(prog="mio", description="", add_help=False)
    parser.add_argument("-h"   , "--help"   , help="Show this help message and exit.", action="store_true", default=False, required=False)
    parser.add_argument("-v"   , "--version", help="Print the mio version and exit." , action="store_true", default=False, required=False)
    parser.add_argument("--dbg",              help="Enable mio tracing output."      , action="store_true", default=False, required=False)
    parser.add_argument("-C"   , "--wd"     , help="Run as if mio was started in <path> instead of the current working directory.", type=pathlib.Path, required=False)
    subparsers = parser.add_subparsers(help='Command to be performed by mio', dest='command')
    
    parser_help = subparsers.add_parser('help', description="Provides documentation on specific command")
    parser_help.add_argument("cmd",  help='Command whose documentation to print', choices=commands)
    
    parser_doctor = subparsers.add_parser('doctor', description=help_text.doctor_help_text, add_help=False)
    
    parser_init = subparsers.add_parser('init', description=help_text.init_help_text, add_help=False)
    
    parser_new = subparsers.add_parser('new', help=help_text.new_help_text, add_help=False)
    parser_new.add_argument('-t', '--template', help='Template name', choices=new.templates, required=False)
    
    parser_install = subparsers.add_parser('install', help=help_text.install_help_text, add_help=False)
    parser_install.add_argument('ip'              , help='Target IP'                                       )
    parser_install.add_argument("-g", "--global"  , help="Install dependencies under '~/.mio'." , action="store_true", default=False, required=False, dest="is_global")
    parser_install.add_argument('-u', "--username", help='Moore.io IP Marketplace username', required=False)
    parser_install.add_argument('-p', "--password", help='Moore.io IP Marketplace password', required=False)
    
    parser_login = subparsers.add_parser('login', help=help_text.login_help_text, add_help=False)
    parser_login.add_argument('-u', "--username", help='Moore.io IP Marketplace username', required=False)
    parser_login.add_argument('-p', "--password", help='Moore.io IP Marketplace password', required=False)
    
    parser_publish = subparsers.add_parser('publish', help=help_text.publish_help_text, add_help=False)
    parser_publish.add_argument('ip'              , help='Target IP'                                       )
    parser_publish.add_argument('-u', "--username", help='Moore.io IP Marketplace username', required=False)
    parser_publish.add_argument('-p', "--password", help='Moore.io IP Marketplace password', required=False)
    parser_publish.add_argument('-o', "--org"     , help='Moore.io IP Marketplace Organization client name.  Commercial IPs only.', required=False)
    
    parser_package = subparsers.add_parser('package', help=help_text.package_help_text, add_help=False)
    parser_package.add_argument('ip'            , help='Target IP'                          )
    parser_package.add_argument('dest'          , help='Destination path', type=pathlib.Path)
    parser_package.add_argument('-n', "--no-tgz", help='Do not create compressed tarball', required=False, action="store_true", dest="no_tarball")
    
    parser_sim = subparsers.add_parser('sim', help=help_text.sim_help_text, add_help=False)
    parser_sim.add_argument('ip'               , help='Target IP'                                                                                                                               )
    parser_sim.add_argument('-t', "--test"     , help='Delete compiled IP dependencies.'                                                                                        , required=False)
    parser_sim.add_argument('-s', "--seed"     , help='Specify the seed for constrained-random testing.  If none is provided, a random one will be picked.', type=int           , required=False)
    parser_sim.add_argument('-v', "--verbosity", help='Specify the UVM verbosity level for logging: none, low, medium, high or debug.  Default: medium'    , choices=uvm_levels , required=False)
    parser_sim.add_argument('-e', "--errors"   , help='Specifies the number of errors at which compilation/elaboration/simulation is terminated.'          , type=int           , required=False)
    parser_sim.add_argument('-a', "--app"      , help='Specifies which simulator to use: viv, mdc, vcs, xcl, qst, riv.'                                    , choices=simulators , required=False)
    parser_sim.add_argument('-w', "--waves"    , help='Enable wave capture to disk.'                                                                       , action="store_true", required=False)
    parser_sim.add_argument('-c', "--cov"      , help='Enable code & functional coverage capture.'                                                         , action="store_true", required=False)
    parser_sim.add_argument('-g', "--gui"      , help="Invoke the simulator's Graphical User Interface."                                                   , action="store_true", required=False)
    #parser_sim.add_argument('-p', "--prism"    , help='Enable Moore.io PRISM advanced UVM debugging.'                                                      , action="store_true", required=False)
    parser_sim.add_argument('-S'               , help='Force mio to simulate target IP.  Can be combined with -F, -C and/or -E.'                           , action="store_true", required=False)
    parser_sim.add_argument('-E'               , help='Force mio to elaborate target IP.  Can be combined with -F, -C and/or -S.'                          , action="store_true", required=False)
    parser_sim.add_argument('-C'               , help='Force mio to compile target IP.  Can be combined with -F, -E and/or -S.'                            , action="store_true", required=False)
    parser_sim.add_argument('-F'               , help='Force mio to invoke FuseSoC on core file(s).  Can be combined with -C, -E and/or -S.'               , action="store_true", required=False)
    parser_sim.add_argument('-+', "--args"     , help='Add arguments for compilation (+define+NAME[=VALUE]) or simulation (+NAME[=VALUE])).', nargs='+'    , dest='add_args'    , required=False)
    
    parser_repeat = subparsers.add_parser('!', help=help_text.repeat_help_text, add_help=False)
    parser_repeat.add_argument("cmd",  help='Command to be repeated', choices=repeat_commands)
    parser_repeat.add_argument('-b', "--bwrap", help='Does not run command, only creates bash script and packages project.', action="store_true", default=False , required=False)
    
    parser_regr = subparsers.add_parser('regr', help=help_text.regr_help_text, add_help=False)
    parser_regr.add_argument('ip'         , help='Target IP')
    parser_regr.add_argument('regr'       , help='Regression to be run.  For Test Bench IPs with multiple Test Suites, the suite must be specified. Ex: `mio regr my_ip apbxc.sanity`')
    parser_regr.add_argument('-a', "--app", help='Specifies which simulator to use: viv, mdc, vcs, xcl, qst, riv.', choices=simulators , required=False)
    parser_regr.add_argument('-d', "--dry", help='Compiles and elaborates target IP but only prints out the tests that would be run.', action="store_true", default=False , required=False)
    
    parser_clean = subparsers.add_parser('clean', help=help_text.clean_help_text, add_help=False)
    parser_clean.add_argument('ip'          , help='Target IP'                                                            )
    #parser_clean.add_argument('-d', "--deep", help='Delete compiled IP external dependencies.', action="store_true", required=False)
    
    parser_results = subparsers.add_parser('results', help=help_text.results_help_text, add_help=False)
    parser_results.add_argument('ip'      , help='Target IP'      )
    parser_results.add_argument('filename', help='Report filename')
    
    parser_cov = subparsers.add_parser('cov', help=help_text.cov_help_text, add_help=False)
    parser_cov.add_argument('ip', help='Target IP')
    
    parser_dox = subparsers.add_parser('dox', help=help_text.dox_help_text, add_help=False)
    parser_dox.add_argument('ip', help='Target IP')
    
    return parser


def create_sim_job(cli_args):
    sim_job = sim.SimulationJob(cli_args.ip.lower())
    sim_job.is_regression = False
    
    if cli_args.app != None:
        cli_args.app = cli_args.app.lower()
        if cli_args.app == "viv":
            sim_job.simulator = common.simulators_enum.VIVADO
        elif cli_args.app == "mdc":
            sim_job.simulator = common.simulators_enum.METRICS
        elif cli_args.app == "vcs":
            sim_job.simulator = common.simulators_enum.VCS
        elif cli_args.app == "xcl":
            sim_job.simulator = common.simulators_enum.XCELIUM
        elif cli_args.app == "qst":
            sim_job.simulator = common.simulators_enum.QUESTA
        elif cli_args.app == "riv":
            sim_job.simulator = common.simulators_enum.RIVIERA
        else:
            common.dbg("Picked default simulator: vivado")
            sim_job.simulator = common.simulators_enum.VIVADO
    else:
        common.dbg("Picked default simulator: vivado")
        sim_job.simulator = common.simulators_enum.VIVADO
    common.dbg("Using simulator " + str(sim_job.simulator))
    
    if not cli_args.F and not cli_args.C and not cli_args.E and not cli_args.S:
        sim_job.fsoc      = True
        sim_job.compile   = True
        sim_job.elaborate = True
        sim_job.simulate  = True
    else:
        if cli_args.F:
            sim_job.fsoc = True
        else:
            sim_job.fsoc = False
        if cli_args.C:
            sim_job.compile = True
        else:
            sim_job.compile = False
        if cli_args.E:
            sim_job.elaborate = True
        else:
            sim_job.elaborate = False
        if cli_args.S:
            sim_job.simulate = True
        else:
            sim_job.simulate = False
    
    if cli_args.seed == None:
        sim_job.seed = random.randint(1, 2147483646)
        common.dbg("Seed not specified, picked " + str(sim_job.seed))
    else:
        sim_job.seed = cli_args.seed
    
    if cli_args.errors == None:
        sim_job.max_errors = 10
    else:
        sim_job.max_errors = cli_args.errors
    
    if cli_args.verbosity == None:
       sim_job.verbosity = "medium"
    else:
       sim_job.verbosity = cli_args.verbosity
    
    sim_job.test     = cli_args.test
    sim_job.raw_args = cli_args.add_args
    sim_job.waves    = cli_args.waves
    sim_job.cov      = cli_args.cov
    sim_job.gui      = cli_args.gui
    
    if cli_args.gui:
        sim_job.waves = False
    
    return sim_job


def create_repeat_sim_job(cli_args):
    last_cli_args = get_last_job()
    parser        = build_parser()
    sim_args      = parser.parse_args(last_cli_args)
    sim_job = create_sim_job(sim_args)
    sim_job.dry_run = cli_args.bwrap
    sim_job.bwrap   = cli_args.bwrap
    return sim_job


def print_help_text():
    print(help_text.main_help_text)
    common.exit(False)


def print_cmd_help_text(cli_args):
    if cli_args.cmd == "clean":
        print(help_text.clean_help_text)
    if cli_args.cmd == "cov":
        print(help_text.cov_help_text)
    if cli_args.cmd == "doctor":
        print(help_text.doctor_help_text)
    if cli_args.cmd == "dox":
        print(help_text.dox_help_text)
    if cli_args.cmd == "init":
        print(help_text.init_help_text)
    if cli_args.cmd == "install":
        print(help_text.install_help_text)
    if cli_args.cmd == "login":
        print(help_text.login_help_text)
    if cli_args.cmd == "new":
        print(help_text.new_help_text)
    if cli_args.cmd == "package":
        print(help_text.package_help_text)
    if cli_args.cmd == "publish":
        print(help_text.publish_help_text)
    if cli_args.cmd == "regr":
        print(help_text.regr_help_text)
    if cli_args.cmd == "results":
        print(help_text.results_help_text)
    if cli_args.cmd == "sim":
        print(help_text.sim_help_text)
    if cli_args.cmd == "!":
        print(help_text.repeat_help_text)


def print_version():
    print(help_text.version_text)
    common.exit(False)


def get_last_job():
    try:
        if not os.path.exists(cfg.commands_file_path):
            common.fatal(f"No command history exists!")
        with open(cfg.commands_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            timestamps = sorted(ymlr)
            common.dbg(str(ymlr[timestamps[-1]]))
            job = ymlr[timestamps[-1]]
            job.pop(0)
            return job
    except Exception as e:
        common.fatal("Failed to load command history from disk: " + str(e))



def log_cli_args_to_disk():
    if "!" in sys.argv:
        return
    if "login" in sys.argv:
        return
    try:
        common.create_file(cfg.commands_file_path)
        with open(cfg.commands_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = {}
            ymlr[common.timestamp()] = sys.argv
            with open(cfg.commands_file_path, 'w') as yaml_file_write:
                yaml.dump(ymlr, yaml_file_write)
    except Exception as e:
        common.warning("Failed to log command history to disk: " + str(e))
